/* $Id: load.c,v 1.2 2000/03/15 03:39:43 ajapted Exp $
******************************************************************************

   LibGFFD - rule loading code

   Copyright (C) 2000  Andrew Apted   [andrew.apted@ggi-project.org]
   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#define GD_DEBUG_PKG    "LibGFFD"
#define GD_DEBUG_LEVEL	0

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "internal.h"


static d_list_t /* of gffd_rule_t */ rule_list;

static char *current_class = NULL;

static int linenum = 0;


#define QUOTE_CHAR  '\''

#define SPECIAL_SYMBOL(c)  \
	(((c) == '[') || ((c) == ']') ||  \
	 ((c) == '(') || ((c) == ')'))

#define MAX_STRING  256
#define MAX_LINE    1024


static int parseLine(FILE *fp, char *buf, int max)
{
  int c = 0;
  int quoting = 0;
  int commenting = 0;

  linenum++;

  for (;;) 
  {
    c = fgetc(fp);

    if (c == EOF)
    {
      *buf = 0;
      return -1;
    }

    if (c == '\n')
      break;

    if (commenting)
      continue;

    if ((c == '#') && ! quoting) 
    {
      commenting = 1;
      continue;
    }

    if (c == QUOTE_CHAR)
      quoting = !quoting;

    if (max > 1) 
    {
      *buf++ = c; max--;
    }
  }

  *buf = 0;

  return 0;
}

static const char *skipSpaces(const char *s)
{
  for (; *s && isspace(*s); s++) 
  { }

  return s;
}

static const char *parseWord(char *dest, int max, const char *src)
{
  src = skipSpaces(src);

  if (*src == 0)
  {
    *dest = 0;
    return src;
  }

  if (SPECIAL_SYMBOL(*src)) 
  {
    *dest++ = *src++;
    *dest = 0;
    return src;
  }

  for (; *src && ! isspace(*src) && ! SPECIAL_SYMBOL(*src); src++) 
  {
    if (max > 1) 
    {
      *dest++ = *src; 
      max--;
    }
  }

  *dest = 0;
  return src;
}

static const char *parseByteString(unsigned char *dest, int *len,
    int max, const char *src)
{
  (*len) = 0;

  for (;;) 
  {
    src = skipSpaces(src);

    if (*src == QUOTE_CHAR) 
    {
      for (src++; *src && *src != QUOTE_CHAR; src++)
      {
        if ((*len) >= max) 
        {
          GD_ERROR("ByteString overflow.\n");
          return NULL;
        }

        *dest++ = *src; (*len)++;
      }

      if (*src != QUOTE_CHAR) 
      {
        GD_ERROR("Missing endquote.\n");
        return NULL;
      }

      src++;
      continue;
    }

    if (isdigit(*src) || (*src == '-'))
    {
      char numbuf[MAX_STRING];
      int val;

      src = parseWord(numbuf, MAX_STRING, src);

      if (sscanf(numbuf, "%i", &val) != 1) 
      {
        GD_ERROR("Bad number: %s\n", numbuf);
        return NULL;
      }

      if ((val < 0) || (val > 255)) 
      {
        GD_ERROR("Illegal byte value: %d\n", val);
        return NULL;
      }

      if ((*len) >= max) 
      {
        GD_ERROR("ByteString overflow.\n");
        return NULL;
      }

      *dest++ = (unsigned char) val; (*len)++;
      continue;
    }

    /* no more byte-stream-ish data */

    return src;
  }

  return NULL;  /* not actually reached */
}


/* ---------------------------------------------------------------------- */


static void freeOneName(d_node_t *N)
{
  gffd_name_t *name = (gffd_name_t *) N;

  free((void *) name->name);

  listFreeNode(N);
}

static void freeOnePart(d_node_t *N)
{
  gffd_part_t *part = (gffd_part_t *) N;

  free((void *) part->data);

  listFreeNode(N);
}

static void freeOneMagic(d_node_t *N)
{
  gffd_magic_t *M = (gffd_magic_t *) N;

  listFree(&M->parts, &freeOnePart);
  listFreeNode(N);
}

static void freeOneRule(d_node_t *N)
{
  gffd_rule_t *R = (gffd_rule_t *) N;

  if (R->class)
    free((void *) R->class);

  if (R->type)
    free((void *) R->type);

  if (R->MIME)
    free((void *) R->MIME);
  
  listFree(&R->names,  &freeOneName);
  listFree(&R->magics, &freeOneMagic);

  listFreeNode(N);
}

static gffd_rule_t *newRule(const char *class)
{
  gffd_rule_t *R;

  R = (gffd_rule_t *) malloc(sizeof(gffd_rule_t));

  R->class = strdup(class);
  R->type  = NULL;
  R->MIME  = NULL;

  listInit(&R->names);
  listInit(&R->magics);

  return R;
}


/* ---------------------------------------------------------------------- */


static int handleClass(const char *line)
{
  char tokbuf[MAX_STRING];

  if (current_class)
  {
    free(current_class);
    current_class = NULL;
  }

  line = parseWord(tokbuf, MAX_STRING, line);
  
  if (tokbuf[0] == 0)
  {
    GD_ERROR("Bad class name.\n");
    return -1;
  }

  current_class = strdup(tokbuf);

  GD_DEBUG2("PARSING CLASS: %s\n", tokbuf);
  return 0;
}

static int handleType(gffd_rule_t *R, const char *line)
{
  char tokbuf[MAX_STRING];

  GD_ASSERT(R);
  GD_ASSERT(! R->type);

  line = parseWord(tokbuf, MAX_STRING, line);
  
  if (tokbuf[0] == 0)
  {
    GD_ERROR("Bad type name.\n");
    return -1;
  }

  R->type = strdup(tokbuf);

  GD_DEBUG2("PARSING TYPE: %s\n", tokbuf);
  return 0;
}

static int handleName(gffd_rule_t *R, const char *line)
{
  gffd_name_t *name;

  char tokbuf[MAX_STRING];

  int len;

  if (! R)
  {
    GD_ERROR("Type not yet specified !\n");
    return -1;
  }

  name = (gffd_name_t *) malloc(sizeof(gffd_name_t));

  line = parseWord(tokbuf, MAX_STRING, line);

  if (sscanf(tokbuf, "%i", &name->score) != 1)
  {
    GD_ERROR("Bad name score.\n", tokbuf);
    free(name);
    return -1;
  }

  if (name->score < 1 || name->score > 3)
  {
    GD_ERROR("Illegal name score value: %d\n", name->score);
    free(name);
    return -1;
  }

  line = parseByteString((unsigned char *)tokbuf, &len, MAX_STRING-1, line);
  tokbuf[len] = 0;

  if (line == NULL || len == 0) 
  {
    GD_ERROR("Bad name string.\n");
    free(name);
    return -1;
  }

  name->name = strdup(tokbuf);

  listAddTail(&R->names, &name->n);

  GD_DEBUG3("Added name `%s' for type %s.\n", name->name, R->type);
  return 0;
}

static int handleParts(gffd_magic_t *magic, const char *line)
{
  gffd_part_t *part;

  char tokbuf[MAX_STRING];

  unsigned char databuf[MAX_STRING];
  unsigned char maskbuf[MAX_STRING];

  int len, i;
  int approx_match;

  for (;;)
  {
    part = (gffd_part_t *) malloc(sizeof(gffd_part_t));

    line = parseWord(tokbuf, MAX_STRING, line);

    if (sscanf(tokbuf, "%i", &part->offset) != 1 ||
        part->offset < 0)
    {
      GD_ERROR("Bad data offset in magic expression.\n");
      free(part);
      return -1;
    }

    line = parseWord(tokbuf, MAX_STRING, line);

    if (strcmp(tokbuf, "[") != 0)
    {
      GD_ERROR("Missing '[' in magic expression.\n");
      free(part);
      return -1;
    }

    line = parseWord(tokbuf, MAX_STRING, line);

    if (sscanf(tokbuf, "%i", &part->length) != 1 ||
        part->length <= 0)
    {
      GD_ERROR("Bad data length in magic expression.\n");
      free(part);
      return -1;
    }

    if (part->length > MAX_STRING)
    {
      GD_ERROR("Data length too long in magic expression.\n");
      free(part);
      return -1;
    }

    line = parseWord(tokbuf, MAX_STRING, line);

    if (strcmp(tokbuf, "]") != 0)
    {
      GD_ERROR("Missing ']' in magic expression.\n");
      free(part);
      return -1;
    }

    memset(maskbuf, 0xff, part->length);

    line = parseWord(tokbuf, MAX_STRING, line);

    if (strcmp(tokbuf, "&") == 0)
    {
      line = parseByteString(maskbuf, &len, MAX_STRING, line);

      if (line == NULL)
      {
        GD_NOTICE("Error parsing mask bytes.\n");
        free(part); 
        return -1;
      }

      if (len < part->length)
      {
        GD_ERROR("Mask data too short in magic expr (%d < %d).\n",
            len, part->length);
        free(part);
        return -1;
      }

      if (len > part->length) {

        GD_ERROR("Mask data too long in magic expr (%d > %d).\n",
            len, part->length);
        free(part);
        return -1;
      }

      line = parseWord(tokbuf, MAX_STRING, line);
    }

    approx_match = 0;

    if (strcmp(tokbuf, "~=") == 0)
      approx_match = 1;

    if (strcmp(tokbuf, "==") != 0 && strcmp(tokbuf, "~=") != 0)
    {
      GD_ERROR("Operator %s unknown.\n", tokbuf);
      free(part);
      return -1;
    }

    line = parseByteString(databuf, &len, MAX_STRING, line);

    if (line == NULL)
    {
      GD_NOTICE("Error parsing comparison bytes.\n");
      free(part); 
      return -1;
    }

    if (len < part->length)
    {
      GD_ERROR("Comparison data too short in magic expr (%d < %d).\n",
            len, part->length);
      free(part);
      return -1;
    }

    if (len > part->length)
    {
      GD_ERROR("Comparison data too long in magic expr (%d > %d).\n",
            len, part->length);
      free(part);
      return -1;
    }

    /* check for bad data, and update data/mask for the ~=
     * case-insensitive operator.
     */

    for (i=0; i < part->length; i++)
    {
      if (approx_match)
      {
        if (isalpha(databuf[i]))
        {
          databuf[i] &= 0xdf;
          maskbuf[i] &= 0xdf;
        }
      }

      if (databuf[i] & ~maskbuf[i])
      {
        GD_ERROR("Comparison data [%d] is outside mask.\n", i);
        free(part);
        return -1;
      }
    }

    /* store values in the magic part */

    part->data = (unsigned char *) malloc(part->length * 2);

    for (i=0; i < part->length; i++)
    {
      part->data[i*2 + 0] = maskbuf[i];
      part->data[i*2 + 1] = databuf[i];
    }
    
    listAddTail(&magic->parts, &part->n);

    /* check for end-of-line or mary-go-around again */

    line = skipSpaces(line);

    if (*line == 0)
    {
      /* THE END */
      return 0;
    }

    line = parseWord(tokbuf, MAX_STRING, line);

    if (strcmp(tokbuf, "&&") != 0)
    {
      GD_ERROR("Expecting `&&' in magic expression.\n");
      return -1;
    }
  }

  return 0;   /* never actually reached */
}

static int handleMagic(gffd_rule_t *R, const char *line)
{
  gffd_magic_t *magic;

  char tokbuf[MAX_STRING];

  if (! R)
  {
    GD_ERROR("Type not yet specified !\n");
    return -1;
  }

  magic = (gffd_magic_t *) malloc(sizeof(gffd_magic_t));

  listInit(&magic->parts);

  line = parseWord(tokbuf, MAX_STRING, line);

  if (sscanf(tokbuf, "%i", &magic->score) != 1)
  {
    GD_ERROR("Bad magic score.\n");
    free(magic);
    return -1;
  }

  if (magic->score < 1 || magic->score > 3)
  {
    GD_ERROR("Illegal magic score value: %d\n", magic->score);
    free(magic);
    return -1;
  }

  if (handleParts(magic, line) < 0)
  {
    free(magic);
    return -1;
  }

  listAddTail(&R->magics, &magic->n);

  GD_DEBUG3("Added magic for type %s.\n", R->type);
  return 0;
}

static int handleMIME(gffd_rule_t *R, const char *line)
{
  char tokbuf[MAX_STRING];

  if (! R)
  {
    GD_ERROR("Type not yet specified !\n");
    return -1;
  }

  if (R->MIME)
  {
    GD_ERROR("Rule already has a mime type: %s\n", R->MIME);
    return -1;
  }

  line = parseWord(tokbuf, MAX_STRING, line);
  
  if (strchr(tokbuf, '/') == NULL)
  {
    GD_ERROR("Bad MIME string.\n");
    return -1;
  }

  /* !!! FIXME: make lower case */
  
  R->MIME = strdup(tokbuf);

  GD_DEBUG3("MIME STRING: %s\n", tokbuf);
  return 0;
}

static int handleFunction(gffd_rule_t *R, const char *line)
{
  if (! R)
  {
    GD_ERROR("Type not yet specified !\n");
    return -1;
  }

  GD_ERROR("Functions not yet implemented.\n");
  return -1;
}


/* ---------------------------------------------------------------------- */


void _gffdInitRules(void)
{
  listInit(&rule_list);
}

int _gffdLoadRules(const char *filename)
{
  FILE *fp;

  gffd_rule_t *cur = NULL;

  char linebuf[MAX_LINE];
  const char *s;

  int errors = 0;


  /* open config file */

  GD_DEBUG1("Loading Rules: started\n");

  fp = fopen(filename, "r");

  if (fp == NULL)
  {
    GD_ERROR("Failed to open file `%s'.\n", filename);
    return GFFD_E_ERROR;
  }

  /* parse each line */

  while (parseLine(fp, linebuf, MAX_LINE) == 0) 
  {
    s = skipSpaces(linebuf);

    if (*s == 0)
      continue;

    if (strncmp(s, "class:", 6) == 0) 
    {
      if (handleClass(skipSpaces(s + 6)) < 0)
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    }

    if (strncmp(s, "type:", 5) == 0) 
    {
      if (cur != NULL)
      {
        listAddTail(&rule_list, &cur->n);
        cur = NULL;
      }

      if (! current_class)
      {
        GD_ERROR("Class not yet specified !\n");
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        errors++;
        continue;
      }

      cur = newRule(current_class);

      if (handleType(cur, skipSpaces(s + 5)) < 0)
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    } 

    if (strncmp(s, "mime:", 5) == 0)
    {
      if (handleMIME(cur, skipSpaces(s + 5)) < 0)
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    } 

    if (strncmp(s, "name:", 5) == 0) 
    {
      if (handleName(cur, skipSpaces(s + 5)) < 0)
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    } 

    if (strncmp(s, "magic:", 6) == 0) 
    {
      if (handleMagic(cur, skipSpaces(s + 6)) < 0) 
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    }

    if (strncmp(s, "func:", 5) == 0) 
    {
      if (handleFunction(cur, skipSpaces(s + 5)) < 0) 
      {
        GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
        GD_NOTICE("Line contains: %s\n\n", s);
        errors++;
      }
      continue;
    }

    GD_ERROR("Unknown keyword in line %d of file `%s'.\n",
        linenum, filename);
    GD_NOTICE("Occurred in line %d of file `%s'.\n", linenum, filename);
    GD_NOTICE("Line contains: %s\n\n", s);
    errors++;
  }

  fclose(fp);

  /* tidy up loose ends */

  if (current_class)
  {
    free(current_class);
    current_class = NULL;
  }

  if (cur != NULL) 
  {
    listAddTail(&rule_list, &cur->n);
    cur = NULL;
  }

  if (errors > 0)
  {
    _gffdFreeRules();
    GD_NOTICE("Encountered %d errors during parsing.\n", errors);
    return GFFD_E_ERROR;
  }

  if (! rule_list.head)
  {
    GD_ERROR("There were no rules in the rulefile !\n");
    return GFFD_E_ERROR;
  }
 
  GD_DEBUG1("Loading Rules: successful !\n");

  return 0;
}

void _gffdFreeRules(void)
{
  listFree(&rule_list, &freeOneRule);
}

d_list_t /* of gffd_rule_t */ * _gffdGetRules(void)
{
  GD_ASSERT(rule_list.head);

  return &rule_list;
}

